from typing import Generic, List, Optional, Type, TypeVar

from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel
from sqlalchemy.orm import Session

from lgbsttracker.store.db.sql_background import run_async
from lgbsttracker.store.db.sql_database import db

ModelType = TypeVar("ModelType", bound=db.BaseModel)  # type: ignore
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class CRUDBase(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, model: Type[ModelType]):
        """
        CRUD object with default methods to Create, Read, Update, Delete (CRUD).

        **Parameters**

        * `model`: A SQLAlchemy model class
        * `schema`: A Pydantic model (schema) class
        """
        self.model = model

    async def get(self, db_session: Session, id: int) -> Optional[ModelType]:
        return await run_async(db_session.query(self.model).filter(self.model.id == id).first)  # type: ignore

    async def get_multi(self, db_session: Session, *, skip=0, limit=100) -> List[ModelType]:
        return await run_async(db_session.query(self.model).offset(skip).limit(limit).all)

    async def create(self, db_session: Session, *, obj_in: CreateSchemaType) -> ModelType:
        def _create(db_session, obj_in):
            obj_in_data = jsonable_encoder(obj_in)
            db_obj = self.model(**obj_in_data)
            db_session.add(db_obj)
            db_session.commit()
            db_session.refresh(db_obj)
            return db_obj

        return await run_async(_create, db_session, obj_in)

    async def update(self, db_session: Session, *, db_obj: ModelType, obj_in: UpdateSchemaType) -> ModelType:
        def _update(db_session, db_obj, obj_in):
            obj_data = jsonable_encoder(db_obj)
            update_data = obj_in.dict(skip_defaults=True)
            for field in obj_data:
                if field in update_data:
                    setattr(db_obj, field, update_data[field])
            db_session.add(db_obj)
            db_session.commit()
            db_session.refresh(db_obj)
            return db_obj

        return await run_async(_update, db_session, db_obj, obj_in)

    async def remove(self, db_session: Session, *, id: int) -> ModelType:
        def _remove(db_session, model, id):
            obj = db_session.query(model).get(id)
            db_session.delete(obj)
            db_session.commit()
            return obj

        return await run_async(_remove, db_session, self.model, id)



from sqlalchemy.ext.declarative import declared_attr


class CustomBase(object):
    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()



from __future__ import print_function

import logging
import logging.config

# Logging format example:
# 2018/11/20 12:36:37 INFO lgbsttracker.event.kafka: Creating Kafka Event
LOGGING_LINE_FORMAT = "%(asctime)s %(levelname)s %(name)s: %(message)s"
LOGGING_DATETIME_FORMAT = "%Y/%m/%d %H:%M:%S"


def _configure_lgbsttracker_loggers(root_module_name, debug_level="INFO", app_name="app"):
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "lgbsttracker_formatter": {"format": LOGGING_LINE_FORMAT, "datefmt": LOGGING_DATETIME_FORMAT}
            },
            "handlers": {
                "console_handler": {
                    "level": "DEBUG",
                    "formatter": "lgbsttracker_formatter",
                    "class": "logging.StreamHandler",
                },
                "log_file_handler": {
                    "level": "DEBUG",
                    "formatter": "lgbsttracker_formatter",
                    "class": "logging.handlers.RotatingFileHandler",
                    "filename": f"{app_name}.log",
                    "maxBytes": 10485760,
                    "backupCount": 20,
                    "encoding": "utf8",
                },
                "error_file_handler": {
                    "level": "ERROR",
                    "formatter": "lgbsttracker_formatter",
                    "class": "logging.handlers.RotatingFileHandler",
                    "filename": f"{app_name}-error.log",
                    "maxBytes": 10485760,
                    "backupCount": 20,
                    "encoding": "utf8",
                },
            },
            "loggers": {
                root_module_name: {
                    "handlers": ["console_handler", "log_file_handler", "error_file_handler"],
                    "level": debug_level,
                    "propagate": False,
                }
            },
        }
    )



from six.moves import urllib

from lgbsttracker.errors import ErrorCodes
from lgbsttracker.exceptions import GenericException
from lgbsttracker.store.db.db_types import DATABASE_ENGINES
from lgbsttracker.utils.validation import _validate_db_type_string

_INVALID_DB_URI_MSG = "Please refer to format specifications."


def extract_db_type_from_uri(db_uri):
    """
    Parse the specified DB URI to extract the database type. Confirm the database type is
    supported. If a driver is specified, confirm it passes a plausible regex.
    """
    scheme = urllib.parse.urlparse(db_uri).scheme
    scheme_plus_count = scheme.count("+")

    if scheme_plus_count == 0:
        db_type = scheme
    elif scheme_plus_count == 1:
        db_type, _ = scheme.split("+")
    else:
        error_msg = "Invalid database URI: '%s'. %s" % (db_uri, _INVALID_DB_URI_MSG)
        raise GenericException(error_msg, ErrorCodes.INVALID_PARAMETER_VALUE)

    _validate_db_type_string(db_type)

    return db_type


def get_uri_scheme(uri_or_path):
    scheme = urllib.parse.urlparse(uri_or_path).scheme
    if any([scheme.lower().startswith(db) for db in DATABASE_ENGINES]):
        return extract_db_type_from_uri(uri_or_path)
    else:
        return scheme


def extract_host_name_from_uri(uri_or_path):
    return urllib.parse.urlparse(uri_or_path).hostname



"""initial db

Revision ID: ae2444ccb82a
Revises: 
Create Date: 2020-01-28 10:51:41.173284

"""
import datetime
import logging

from alembic import op
from sqlalchemy import CheckConstraint, Column, DateTime, Float, Index, Integer, PrimaryKeyConstraint, String, orm

from lgbsttracker.entities.experiment import ExperimentAction, ExperimentState

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)

# revision identifiers, used by Alembic.
revision = "ae2444ccb82a"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    bind = op.get_bind()
    session = orm.Session(bind=bind)

    _logger.info("Initial database creation.")

    # Experiment table
    op.create_table(
        "experiment",
        Column("id", Integer, autoincrement=True),
        Column("experiment_uuid", String(32), nullable=False),
        Column("ts", DateTime, default=str(datetime.datetime.now())),
        Column("action", String(32), default=ExperimentAction.NONE),
        Column("vision_sensor", Float, nullable=True),
        Column("speed", Float, nullable=True),
        Column("state", String(32), default=ExperimentState.STOPPED),
        # Primary Key
        PrimaryKeyConstraint("id", name="experiment_pk"),
        # Check Constraint
        CheckConstraint("action IN ('none', 'forward', 'backward', 'left', 'right')", name="experiment_checkaction",),
        CheckConstraint("state IN ('stopped', 'failed', 'running', 'finished')", name="experiment_checkstate",),
        # Index creation
        Index("experiment_index1", "experiment_uuid"),
    )

    session.commit()

    _logger.info("Database creation completed.")


def downgrade():
    pass



import click

import lgbsttracker.store.db.utils


@click.group("db")
def commands():
    """
    Commands for managing databases.
    """


@commands.command()
@click.argument("url", required=False)
def upgrade_db(url):
    """
    Upgrade database to the latest supported version.

    :param url of database
    """
    from lgbsttracker.services.store._experiment_registry import utils

    url = url if url is not None else utils.get_experiment_uri()
    lgbsttracker.store.db.utils.upgrade_sql_db(url)



from abc import ABCMeta, abstractmethod
from typing import List

from lgbsttracker.entities import ExperimentCreate
from lgbsttracker.store.experiment.dbmodels.experiment import Experiment


class AbstractStore:
    """
    Abstract class for Experiment.
    This class defines the API interface for front ends to connect with various types of stores.
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        """
        Empty constructor for now. This is deliberately not marked as abstract, else every
        derived class would be forced to create one.
        """

    @abstractmethod
    def open(self):
        """
        Open Store
        """

    @abstractmethod
    def close(self):
        """
        Close Store
        """

    @abstractmethod
    async def create_experiment(self, entity: ExperimentCreate) -> Experiment:
        """

        :param A :py:class:`lgbsttracker.entity.ExperimentCreate` object.

        :return: A :py:class:`lgbsttracker.entity.Experiment` object.
        """

    @abstractmethod
    async def get_experiment_by_uuid(self, uuid: str) -> List[Experiment]:
        """

        :param id: Experiment Id

        :return: A list of :py:class:`lgbsttracker.entity.Experiment` objects.
        """



from __future__ import print_function

from lgbsttracker.errors import ErrorCodes
from lgbsttracker.exceptions import GenericException
from lgbsttracker.services.store._experiment_registry.registry import ExperimentStoreRegistry
from lgbsttracker.utils import env

_EXPERIMENT_URI_ENV_VAR = "EXPERIMENT_STORAGE_URI"

_experiment_uri = None


def is_experiment_uri_set():
    """Returns True if the experiment URI has been set, False otherwise."""
    if _experiment_uri or env.get_env(_EXPERIMENT_URI_ENV_VAR):
        return True
    return False


def set_experiment_uri(uri):
    """
    Set the Experiment URI storage.
    :param uri: Input parameter string
    """
    global _experiment_uri
    _experiment_uri = uri


def get_experiment_uri():
    """
    Get the Experiment URI storage.

    :return: The Experiment URI.
    """
    global _experiment_uri
    if _experiment_uri is not None:
        return _experiment_uri
    elif env.get_env(_EXPERIMENT_URI_ENV_VAR) is not None:
        return env.get_env(_EXPERIMENT_URI_ENV_VAR)
    else:
        raise GenericException("Storage Experiment URI store is not set", error=ErrorCodes.INTERNAL_ERROR)


def _get_sqlalchemy_store(experiment_uri):
    from lgbsttracker.store.experiment.sqlachemy_store import SqlAlchemyStore

    return SqlAlchemyStore(db_uri=experiment_uri)


_experiment_store_registry = ExperimentStoreRegistry()
_experiment_store_registry.register("sqlite", _get_sqlalchemy_store)
_experiment_store_registry.register("postgresql", _get_sqlalchemy_store)

_experiment_store_registry.register_entrypoints()


def _get_store(experiment_uri=None):
    return _experiment_store_registry.get_store(experiment_uri)



"""
Internal package providing a Python CRUD interface.
"""
from typing import List

from lgbsttracker.entities import Experiment, ExperimentCreate
from lgbsttracker.services.store._experiment_registry import utils


class ExperimentService(object):
    """
    Experiment Service.
    """

    def __init__(self, experiment_uri: str) -> None:
        """
        :param experiment_uri: Address of local or remote experiment db.
        """
        self.experiment_uri = experiment_uri
        self.store = utils._get_store(self.experiment_uri)

    def open(self) -> None:
        """
        Open DB
        """
        self.store.open()

    def close(self) -> None:
        """
        Close DB
        """
        self.store.close()

    async def create_experiment(self, entity: ExperimentCreate,) -> Experiment:
        """
        Create experiment entity.

        :param entity: :py:class:`lgbsttracker.entities.ExperimentCreate` entity

        :return: None
        """
        return await self.store.create_experiment(entity)

    async def get_experiment_by_uuid(self, uuid: str) -> List[Experiment]:
        """
        Get Experiment entry by name.
        :param uuid: Experiment UUID.

        :return: A list :py:class:`lgbsttracker.entities.Experiment` entities
        """
        return await self.store.get_experiment_by_uuid(uuid)



import click

import lgbsttracker.db
from lgbsttracker import __version__


@click.group()
@click.version_option(__version__)
def cli():
    pass


cli.add_command(lgbsttracker.db.commands)

if __name__ == "__main__":
    cli()



from datetime import datetime

from pydantic import BaseModel


class ExperimentAction:
    NONE = "none"
    FORWARD = "forward"
    BACKWARD = "backward"
    LEFT = "left"
    RIGHT = "right"


class ExperimentState:
    STOPPED = "stopped"
    FAILED = "failed"
    RUNNING = "running"
    FINISHED = "finished"


# Shared properties
class ExperimentBase(BaseModel):
    experiment_uuid: str = None  # type: ignore
    ts: datetime = None  # type: ignore
    action: str = None  # type: ignore
    vision_sensor: float = None  # type: ignore
    speed: float = None  # type: ignore
    state: str = None  # type: ignore


# Properties to receive on experiment creation
class ExperimentCreate(ExperimentBase):
    experiment_uuid: str


# Properties to receive on item update
class ExperimentUpdate(ExperimentBase):
    pass


# Properties shared by models stored in DB
class ExperimentInDBBase(ExperimentBase):
    id: int

    class Config:
        orm_mode = True


# Properties to return to client
class Experiment(ExperimentInDBBase):
    pass


# Properties properties stored in DB
class ExperimentInDB(ExperimentInDBBase):
    pass



"""
Database module.
Defines types, functions and primitives for the initialization, disposition and management of
  database connections and their sessions.
"""
from typing import Callable

from sqlalchemy import create_engine as _sa_create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session as SQLAlchemySession, scoped_session, sessionmaker
from sqlalchemy.schema import MetaData

from lgbsttracker.exceptions import DatabaseNotInitialized
from lgbsttracker.store.db.sql_background import run_async
from lgbsttracker.store.db.sql_base import CustomBase
from lgbsttracker.utils import env

__all__ = [
    "transactional_session",
    "DatabaseManager",
    "db",
]

schema = env.get_env("SCHEMA", "public")
min_connection_pool_size = env.get_env("MIN_CONNECTION_POOL_SIZE", 10)
max_connection_pool_size = env.get_env("MAX_CONNECTION_POOL_SIZE", 20)
connection_pool_recycle_time = env.get_env("CONNECTION_POOL_RECYCLE_TIME", 30)


###
# Context Manager for Transactional Session Management
#
class transactional_session:
    """
    Context manager which provides transactional session management.
    Supports the sync/async context manager protocols.
    """

    def __init__(self, expire_on_commit: bool = True):
        """
        Initializer.
        :param bool expire_on_commit: If True, will make a session the expires all objects in memory after commit;
            otherwise, will make objects in memory accessible even after session.commit() is called.
        """
        if db.Session is None or db.OnCommitExpiringSession is None:
            raise DatabaseNotInitialized(message="The global database.db singleton is not initialized!")

        self._session = None
        self._expire_on_commit = expire_on_commit

    ###
    # Synchronous context manager protocol
    #
    def __enter__(self) -> SQLAlchemySession:
        if self._expire_on_commit is True:
            self._session = db.OnCommitExpiringSession()
        else:
            self._session = db.Session()

        return self._session  # type: ignore

    def __exit__(self, exc_type, exc, tb):
        try:
            self._session.commit()
        except Exception as e:
            self._session.rollback()
            _logger.error(e)
            raise e
        finally:
            self._session.close()

    ###
    # Asynchronous context manager protocol
    #
    async def __aenter__(self) -> SQLAlchemySession:
        if self._expire_on_commit is True:
            self._session = db.OnCommitExpiringSession()
        else:
            self._session = db.Session()

        return self._session  # type: ignore

    async def __aexit__(self, exc_type, exc, tb):
        try:
            await run_async(self._session.commit)
        except Exception as e:
            await run_async(self._session.rollback)
            _logger.error(e)
            raise e
        finally:
            self._session.close()


###
# Database Manager
#
class DatabaseManager:
    """
    Configuration class for DB that encapsulates engine and configured class for creating scoped session instances.
    """

    def __init__(self):
        ###
        # Private database engine and metadata attributes.
        #
        self._engine = None
        self._metadata = MetaData(schema=schema)

        ###
        # Session Factory classes, later initialized in self.initialize() method.
        #
        # The self.Session corresponds to a session factory that doesn't expire ORM instances from memory
        #   after getting committed.
        #
        # The self.OnCommitExpiringSession corresponds to a session factory that expires ORM instances from
        #   memory after getting committed.
        #
        self.Session = None
        self.OnCommitExpiringSession = None

        ###
        # Declarative Base Model class.
        #
        self.BaseModel = declarative_base(cls=CustomBase, metadata=MetaData(schema=schema))

    @property
    def engine(self) -> Engine:
        return self._engine

    @classmethod
    def create_database_engine(cls, db_uri: str) -> Engine:
        """
        Creates a new SQLAlchemy database engine (sqlalchemy.engine.base.Engine) and returns it.
        :return: a working SQLAlchemy database engine
        """
        ###
        # Database configuration options
        #
        # # Database connection pool settings
        min_pool_size = min_connection_pool_size
        max_pool_size = max_connection_pool_size

        if max_pool_size < min_pool_size:
            raise ValueError("Max Pool Size cannot be lower than Min Pool Size!")

        max_pool_overflow = max_pool_size - min_pool_size
        pool_recycle_time = connection_pool_recycle_time

        return _sa_create_engine(
            db_uri,
            encoding="utf-8",
            pool_size=min_pool_size,
            max_overflow=max_pool_overflow,
            pool_recycle=pool_recycle_time,
        )

    def initialize(self, db_uri: str, db_engine: Engine = None, scope_function: Callable = None):
        """
        Configure class for creating scoped sessions.
        :param db_engine: DB connection engine.
        :param scope_function: a function for scoping database connections.
        """
        # Set or initialize the database engine
        if db_engine is None:
            self._engine = self.create_database_engine(db_uri)
        else:
            self._engine = db_engine

        # Create the session factory classes
        self.Session = scoped_session(sessionmaker(bind=self._engine, expire_on_commit=False), scopefunc=scope_function)

        self.OnCommitExpiringSession = scoped_session(
            sessionmaker(bind=self._engine, expire_on_commit=True), scopefunc=scope_function
        )

    def cleanup(self):
        """
        Cleans up the database connection pool.
        """
        if self._engine is not None:
            self._engine.dispose()


###
# Database Extension
#
db = DatabaseManager()



from lgbsttracker.errors import ErrorCodeNames, ErrorCodes


class GenericException(Exception):
    """
    Generic exception thrown to surface failure information about external-facing operations.
    The error message associated with this exception may be exposed to clients in HTTP responses
    for debugging purposes. If the error text is sensitive, raise a generic `Exception` object
    instead.
    """

    def __init__(self, message, error_code=ErrorCodes.INTERNAL_ERROR, **kwargs):
        """
        :param message: The message describing the error that occured. This will be included in the
                        exception's serialized JSON representation.
        :param error_code: An appropriate error code for the error that occured; it will be included
                           in the exception's serialized JSON representation. This should be one of
                           the codes listed in the `lgbsttracker.errors.ErrorCodes`.
        :param kwargs: Additional key-value pairs to include in the serialized JSON representation
                       of the GenericException.
        """
        try:
            self.error_code = ErrorCodeNames.get(error_code)
        except (ValueError, TypeError):
            self.error_code = ErrorCodeNames.get(ErrorCodes.INTERNAL_ERROR)
        self.message = message
        self.json_kwargs = kwargs
        super(GenericException, self).__init__(message)


class DatabaseNotInitialized(GenericException):
    def __init__(self, message):
        super(DatabaseNotInitialized, self).__init__(message, error_code=ErrorCodes.INTERNAL_ERROR)



"""
The Background Common module.
Defines primitives for executing I/O blocking operations in the background using
a thread pool executor.
"""
import asyncio
import logging
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Callable

from lgbsttracker.utils import env

__all__ = [
    "run_async",
]


_logger = logging.getLogger()

db_thread_pool_executor_size = env.get_env("SQL_DB_THREAD_POOL_EXECUTOR_SIZE", 100)

###
# Thread pool executor used strictly to submit queries/operations to the database
# in background.
#
DB_THREAD_POOL_EXECUTOR = ThreadPoolExecutor(max_workers=db_thread_pool_executor_size, thread_name_prefix="db_worker_")


async def run_async(func: Callable[..., Any], *args, **kwargs) -> Any:
    """
    Runs a callable on the database thread pool executor using the current thread's
      I/O loop instance.
    Usage:
        def blocking_task(arg1, arg2, arg3):
            # ...
            # does some blocking stuff
            # ...
        run_async(blocking_task, arg1, arg2, arg3)
    :param Callable[..., Any] func: the function or callable to run in background
    :param tuple args: function arguments
    :param dict kwargs: function named arguments
    :return: func's return value
    """
    loop = asyncio.get_event_loop()

    def work():
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()

        if end - start > 5:
            try:
                query = args[0]
            except IndexError:
                query = args
            duration = "%.2f" % (end - start)
            _logger.warn("slow_query_logger: {duration}s for executing {query}".format(duration=duration, query=query))
        return result

    return await loop.run_in_executor(DB_THREAD_POOL_EXECUTOR, work)



from lgbsttracker.errors import ErrorCodes
from lgbsttracker.exceptions import GenericException
from lgbsttracker.store.db.db_types import DATABASE_ENGINES

_UNSUPPORTED_DB_TYPE_MSG = f"Supported database engines are {DATABASE_ENGINES}"


def _validate_db_type_string(db_type):
    """validates db_type parsed from DB URI is supported"""
    if db_type not in DATABASE_ENGINES:
        error_msg = "Invalid database engine: '%s'. '%s'" % (db_type, _UNSUPPORTED_DB_TYPE_MSG)
        raise GenericException(error_msg, ErrorCodes.INVALID_PARAMETER_VALUE)



import logging
import os

import sqlalchemy
import sqlalchemy_utils

_logger = logging.getLogger(__name__)


def _create_sqlalchemy_engine(db_uri):
    return sqlalchemy.create_engine(db_uri)


def _get_package_dir():
    """Returns directory containing python package."""
    current_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.normpath(os.path.join(current_dir, os.pardir, os.pardir))


def _get_alembic_config(db_url, alembic_dir=None):
    """
    Constructs an alembic Config object referencing the specified database and migration script
    directory.

    :param db_url Database URL, like sqlite:///<absolute-path-to-local-db-file>. See
    https://docs.sqlalchemy.org/en/13/core/engines.html#database-urls for a full list of valid
    database URLs.
    :param alembic_dir Path to migration script directory. Uses canonical migration script
    directory under mlflow/alembic if unspecified. TODO: remove this argument in MLflow 1.1, as
    it's only used to run special migrations for pre-1.0 users to remove duplicate constraint
    names.
    """
    from alembic.config import Config

    final_alembic_dir = (
        os.path.join(_get_package_dir(), "store", "db_migrations") if alembic_dir is None else alembic_dir
    )
    config = Config(os.path.join(final_alembic_dir, "alembic.ini"))
    config.set_main_option("script_location", final_alembic_dir)
    config.set_main_option("sqlalchemy.url", db_url)
    return config


def upgrade_sql_db(url):
    """
    Upgrade Model DB and create DataBase if not exists

    :param url Database URL
    """
    # alembic adds significant import time, so we import it lazily
    from alembic import command

    engine = _create_sqlalchemy_engine(url)
    if not sqlalchemy_utils.database_exists(engine.url):
        sqlalchemy_utils.create_database(engine.url)
    _logger.info("Updating database tables at %s", url)
    config = _get_alembic_config(url)
    command.upgrade(config, "heads")



from lgbsttracker.services.store.registry import StoreRegistry


class ExperimentStoreRegistry(StoreRegistry):
    """Scheme-based registry for experiment store implementations
    """

    def __init__(self):
        super(ExperimentStoreRegistry, self).__init__("lgbsttracker.experiment_store")

    def get_store(self, experiment_uri=None):
        """Get a store from the registry based on the scheme of experiment_uri

        :param uri: The store URI. If None, it will be inferred from the environment. This URI
                          is used to select which store implementation to instantiate and
                          is passed to the constructor of the implementation.

        :return: An instance of `lgbsttracker.services.store.experiment.AbstractStore` that fulfills the store URI
                 requirements.
        """
        from lgbsttracker.services.store._experiment_registry import utils

        experiment_uri = experiment_uri if experiment_uri is not None else utils.get_experiment_uri()
        builder = self.get_store_builder(experiment_uri)
        return builder(experiment_uri=experiment_uri)



import os


def get_env(variable_name, default=None):
    return os.environ.get(variable_name, default)


def unset_variable(variable_name):
    if variable_name in os.environ:
        del os.environ[variable_name]



from datetime import datetime

from sqlalchemy import CheckConstraint, Column, DateTime, Float, Index, Integer, PrimaryKeyConstraint, String

from lgbsttracker.entities.experiment import ExperimentAction, ExperimentState
from lgbsttracker.store.db.sql_database import db

ActionTypes = [
    ExperimentAction.NONE,
    ExperimentAction.FORWARD,
    ExperimentAction.BACKWARD,
    ExperimentAction.LEFT,
    ExperimentAction.RIGHT,
]

StateTypes = [
    ExperimentState.STOPPED,
    ExperimentState.FAILED,
    ExperimentState.RUNNING,
    ExperimentState.FINISHED,
]


class Experiment(db.BaseModel):  # type: ignore
    id = Column(Integer, autoincrement=True)
    experiment_uuid = Column(String(32), nullable=False)
    ts = Column(DateTime, default=str(datetime.now()))
    action = Column(String(32), default="none")
    vision_sensor = Column(Float, nullable=True)
    speed = Column(Float, nullable=True)
    state = Column(String(32), default="stopped")

    __table_args__ = (
        PrimaryKeyConstraint("id", name="experiment_pk"),
        CheckConstraint(action.in_(ActionTypes), name="experiment_checkaction"),
        CheckConstraint(state.in_(StateTypes), name="experiment_checkstate"),
        Index("experiment_index1", "experiment_uuid"),
    )

    def __repr__(self):
        return "<Experiment ({}, {}, {},{}, {},{}, {})>".format(
            self.id, self.experiment_uuid, self.ts, self.action, self.vision_sensor, self.speed, self.state
        )



import logging

from lgbsttracker.entities import Experiment, ExperimentCreate
from lgbsttracker.services.store._experiment_registry.service import ExperimentService

# TODO: rework using logger
_logger = logging.getLogger(__name__)

_experiment_service = None


def _get_experiment_service(uri=None):
    from lgbsttracker.services.store._experiment_registry import utils

    global _experiment_service
    if _experiment_service is None:
        experiment_uri = uri or utils.get_experiment_uri()
        _experiment_service = ExperimentService(experiment_uri)
        _experiment_service.open()
    return _experiment_service


async def create_experiment(entity: ExperimentCreate) -> Experiment:
    return await _get_experiment_service().create_experiment(entity)


async def get_experiment_by_uuid(uuid: str) -> Experiment:
    return await _get_experiment_service().get_experiment_by_uuid(uuid)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lgbsttracker',
 'lgbsttracker.entities',
 'lgbsttracker.services',
 'lgbsttracker.services.store',
 'lgbsttracker.services.store._experiment_registry',
 'lgbsttracker.store',
 'lgbsttracker.store.db',
 'lgbsttracker.store.db_migrations',
 'lgbsttracker.store.db_migrations.versions',
 'lgbsttracker.store.experiment',
 'lgbsttracker.store.experiment.dbmodels',
 'lgbsttracker.utils']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp',
 'alembic>=1.3.3,<2.0.0',
 'asynctest>=0.13.0,<0.14.0',
 'autoflake>=1.3.1,<2.0.0',
 'click',
 'entrypoints',
 'fastapi>=0.48.0,<0.49.0',
 'google-api',
 'google-api-core',
 'grpcio',
 'grpcio-tools',
 'lock>=2018.3.25,<2019.0.0',
 'numpy',
 'pandas>=0.25.3,<0.26.0',
 'prometheus-client',
 'protoc_gen_swagger',
 'psycopg2',
 'pydantic>=1.4,<2.0',
 'pytest-aiohttp>=0.3.0,<0.4.0',
 'pytest-asyncio>=0.10.0,<0.11.0',
 'python-dateutil',
 'querystring_parser',
 'requests',
 'sqlalchemy',
 'sqlalchemy-utils']

entry_points = \
{'console_scripts': ['lgbsttracker = lgbsttracker.cli:cli']}

setup_kwargs = {
    'name': 'lgbsttracker',
    'version': '0.0.2',
    'description': '0.0.2',
    'long_description': '<H1>lgbsttracker</H1>\n<p align="center">\n<img src="https://img.shields.io/github/last-commit/py4mac/lgbsttracker.svg">\n<a href="https://github.com/py4mac/" target="_blank">\n    <img src="https://github.com/py4mac/lgbsttracker/workflows/Test/badge.svg" alt="Test">\n</a>\n<a href="https://codecov.io/gh/py4mac/lgbsttracker" target="_blank">\n    <img src="https://codecov.io/gh/py4mac/lgbsttracker/branch/master/graph/badge.svg" alt="Coverage">\n</a>\n<a href="https://pypi.org/project/lgbsttracker" target="_blank">\n    <img src="https://badge.fury.io/py/lgbsttracker.svg" alt="Package version">\n</a>\n</p>\n\n---\n\n**Trello Plan**: <a href="https://trello.com/b/lGICXLqL/lgbsttracker" target="_blank">https://trello.com/b/lGICXLqL/lgbsttracker</a>\n\n**Documentation**: <a href="https://py4mac.github.io/lgbsttracker" target="_blank">https://py4mac.github.io/lgbsttracker</a>\n\n**Source Code**: <a href="https://github.com/py4mac/lgbsttracker" target="_blank">https://github.com/py4mac/lgbsttracker</a>\n\n---\n\n## Requirements\n\nPython 3.6+\n\n\n## Installation\n\n```bash\npip install lgbsttracker\n```\n\n## Example\n\n### Use it\n\n\n(Work in progress)\n\n## Environment variables\n\n| Variable Name |  Description | Default value |\n| --- | --- | --- |\n| EXPERIMENT_STORAGE_URI | DB Experiment Storage URI | Mandatory: to be set by the user before using library |\n| SQL_DB_THREAD_POOL_EXECUTOR_SIZE | Thread pool size | 100 |\n| MIN_CONNECTION_POOL_SIZE | Minimum pool size connection | 10 |\n| MAX_CONNECTION_POOL_SIZE | Maxiumum pool size connection | 20 |\n| CONNECTION_POOL_RECYCLE_TIME | Pool connection recycle time | 30 |\n\n## References\n\n* https://github.com/mlflow/mlflow/\n* https://github.com/undera/pylgbst\n* https://github.com/tiangolo/fastapi\n* https://github.com/aalhour/cookiecutter-aiohttp-sqlalchemy/\n\n## License\n\nThis project is licensed under the terms of the MIT license.',
    'author': 'Py4mac',
    'author_email': 'boisbu@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/py4mac/lgbsttracker',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# pylint: disable=wrong-import-position
import lgbsttracker.services  # noqa
from lgbsttracker.utils.logging_utils import _configure_lgbsttracker_loggers
from lgbsttracker.version import VERSION as __version__  # noqa

_configure_lgbsttracker_loggers(root_module_name=__name__)


create_experiment = lgbsttracker.services.create_experiment
get_experiment_by_uuid = lgbsttracker.services.get_experiment_by_uuid

__all__ = [
    "create_experiment",
    "get_experiment_by_uuid",
]



class ErrorCodes:
    INTERNAL_ERROR = 0
    TEMPORARILY_UNAVAILABLE = 1
    ENDPOINT_NOT_FOUND = 2
    PERMISSION_DENIED = 3
    REQUEST_LIMIT_EXCEEDED = 4
    BAD_REQUEST = 5
    INVALID_PARAMETER_VALUE = 6
    RESOURCE_DOES_NOT_EXIST = 7
    INVALID_STATE = 8
    RESOURCE_ALREADY_EXISTS = 9


ErrorCodeNames = {
    ErrorCodes.INTERNAL_ERROR: "Internal Error",
    ErrorCodes.TEMPORARILY_UNAVAILABLE: "Temporarily Unavailable",
    ErrorCodes.ENDPOINT_NOT_FOUND: "Endpoint not found",
    ErrorCodes.PERMISSION_DENIED: "Permission denied",
    ErrorCodes.REQUEST_LIMIT_EXCEEDED: "Request limit exceeded",
    ErrorCodes.BAD_REQUEST: "Bad request",
    ErrorCodes.INVALID_PARAMETER_VALUE: "Invalid parameter value",
    ErrorCodes.RESOURCE_DOES_NOT_EXIST: "Resource does not exist",
    ErrorCodes.INVALID_STATE: "Invalid State",
    ErrorCodes.RESOURCE_ALREADY_EXISTS: "Resource already exists",
}



import logging
from typing import List

from sqlalchemy.orm import Session

from lgbsttracker.entities.experiment import ExperimentCreate, ExperimentUpdate
from lgbsttracker.store.db.sql_background import run_async
from lgbsttracker.store.db.sql_crud_base import CRUDBase
from lgbsttracker.store.db.sql_database import db, schema, transactional_session
from lgbsttracker.store.experiment.abstract_store import AbstractStore
from lgbsttracker.store.experiment.dbmodels.experiment import Experiment

_logger = logging.getLogger()


class CRUDExperiment(CRUDBase[Experiment, ExperimentCreate, ExperimentUpdate]):
    async def get_by_uuid(self, db_session: Session, *, uuid: str) -> List[Experiment]:
        def _get_by_uuid(db_session: Session, uuid: str):
            ret = db_session.query(Experiment).filter(Experiment.experiment_uuid == uuid).all()
            # Hard trick here as the session could expire here.
            db_obj = [
                Experiment(
                    id=e.id, experiment_uuid=e.experiment_uuid, ts=e.ts, action=e.action, vision_sensor=e.vision_sensor, speed=e.speed, state=e.state,
                )
                for e in ret
            ]
            return db_obj

        return await run_async(_get_by_uuid, db_session, uuid)

    async def create(self, db_session: Session, *, obj_in: ExperimentCreate) -> Experiment:
        def _create(db_session: Session, obj_in: ExperimentCreate):
            db_obj = Experiment(experiment_uuid=obj_in.experiment_uuid)
            db_session.add(db_obj)
            db_session.commit()
            db_session.refresh(db_obj)
            return db_obj

        return await run_async(_create, db_session, obj_in)


experiment = CRUDExperiment(Experiment)


class SqlAlchemyStore(AbstractStore):
    def __init__(self, db_uri):
        super(SqlAlchemyStore, self).__init__()
        self.db_uri = db_uri

    def open(self):
        db.initialize(self.db_uri)
        # Check if the table exist
        if not db.engine.dialect.has_table(db.engine, Experiment.__tablename__, schema):
            _logger.info("Creating 'experiment' database")
            Experiment.__table__.create(bind=db.engine)

    def close(self):
        db.cleanup()

    async def create_experiment(self, entity: ExperimentCreate) -> Experiment:
        async with transactional_session() as session:
            _experiment = await experiment.create(session, obj_in=entity)
            return _experiment

    async def get_experiment_by_uuid(self, uuid: str) -> List[Experiment]:
        async with transactional_session() as session:
            _experiments = await experiment.get_by_uuid(session, uuid=uuid)
            return _experiments
